import torch
import torch.nn.functional as F
import torch.nn as nn
import torch.optim as optim

from .event import Event

class TrainTest:
    class Callbacks:
        def __init__(self):
            self.before_start = Event()
            self.before_batch = Event()
            self.before_epoch = Event()

            self.after_end = Event()
            self.after_batch = Event()
            self.after_epoch = Event()

    def get_optimizer(self, name, model, config):
        if name=='sgd':
            return optim.SGD(model.parameters(), config.train_config.lr, config.train_config.momentum)
        else:
            raise ValueError('Optimizer named {} is not supported'.format(name))

    def __init__(self, model, config, 
                 optimizer, loss_module, initializer):

        self.model = model
        self.train_callbacks = TrainTest.Callbacks()
        self.test_callbacks = TrainTest.Callbacks()

        self.train_device = config.train_config.device
        self.test_device = config.test_config.device

        if isinstance(optimizer, str):
            optimizer = self.get_optimizer(optimizer, model, config)

        if initializer is not None:
            model.apply(initializer)

        self.optimizer = optimizer or optim.SGD(model.parameters(), config.train_config.lr, config.train_config.momentum)
        self.loss_module = loss_module or torch.nn.NLLLoss()

        if self.train_device == self.test_device:
            self.model.to(self.train_device)

    def train_epoch(self, train_loader):
        # if train/test device is not same then we need to move model to different device each time
        if self.train_device != self.test_device:
            self.model.to(self.train_device)

        self.model.train()

        self.train_callbacks.before_epoch.notify(self, train_loader.dataset)

        for input, label in train_loader:
            self.train_callbacks.before_batch.notify(self, input, label)
            input, label = input.to(self.train_device), label.to(self.train_device)
            self.optimizer.zero_grad()
            output = self.model(input)
            loss = self.loss_module(output, label) #default reduction is averaging loss over each sample
            loss.backward()
            self.optimizer.step()

            #self.model.eval() # disable layers like dropout
            #with torch.no_grad():
            #    output = self.model(input)
            #self.model.train()

            self.train_callbacks.after_batch.notify(self, input, label, output, loss.item())

        self.train_callbacks.after_epoch.notify(self, train_loader.dataset)

    def test_epoch(self, test_loader):
        if self.train_device != self.test_device:
            self.model.to(self.test_device)
        self.model.eval()

        self.test_callbacks.before_epoch.notify(self, test_loader.dataset)

        with torch.no_grad():
            for input, label in test_loader:
                self.test_callbacks.before_batch.notify(self, input, label)
                input, label = input.to(self.test_device), label.to(self.test_device)
                output = self.model(input)
                loss = self.loss_module(output, label)
                self.test_callbacks.after_batch.notify(self, input, label, output, loss.item())
                
        self.test_callbacks.after_epoch.notify(self, test_loader.dataset)

    def train_model(self, epochs, train_loader, test_loader):
        self.train_callbacks.before_start.notify(self, epochs, train_loader)
        self.test_callbacks.before_start.notify(self, epochs, test_loader)

        for epoch in range(epochs):
            self.train_epoch(train_loader)
            self.test_epoch(test_loader)

        self.train_callbacks.after_end.notify(self, epochs, train_loader)
        self.test_callbacks.after_end.notify(self, epochs, test_loader)



from collections import OrderedDict

class Event:
    def __init__(self):
        self._listners = OrderedDict()

    def subscribe(self, f):
        self._listners[f] = None

    def unsubscribe(self, f):
        self._listners.pop(f, None)

    def notify(self, *args,**kwargs):
        for f in self._listners.keys():
            f(*args,**kwargs) 



from tensorboardX import SummaryWriter
import torch.nn as nn
from torchvision import utils as tvutils

class DebugProbe:
    class LogSettings:
        model_graph = False
        false_preds = False
        param_histo_freq = 0
        param_histo_next_epoch = 0
        basic = False

    def __init__(self, exp_name, run_name, epoch_config, model, callbacks, metrics, log_dir='d:/tlogs/', log_settings = LogSettings()):
        self.log_writer = SummaryWriter(log_dir + exp_name + '/' + run_name)
        self.epoch_config = epoch_config
        self.run_name = run_name
        self.log_settings = log_settings
        self.metrics = metrics
        self.model = model
        
        callbacks.before_start.subscribe(self.on_before_start)
        callbacks.after_epoch.subscribe(self.on_after_epoch)
        callbacks.after_batch.subscribe(self.on_after_batch)

    def on_before_start(self, train_test, epochs, loader):
        print("[{}] CUDA: {}, Batch Size: {}, Data Size: {}, Batches: {}".format(self.run_name, 
            self.epoch_config.use_cuda, self.epoch_config.batch_size, len(loader.dataset), len(loader)))

    def on_after_batch(self, train_test, input, label, output, loss):
        # dump model diagram on first batch
        if self.log_settings.model_graph and self.metrics.metrics['batch_count'] == 1:
            self.log_writer.add_graph(self.model, input, verbose = True)

        # log false predictions
        if self.log_settings.false_preds:
            pred = output.max(1, keepdim=True)[1] # get the index of the max log-probability
            pairs = zip(input, pred.eq(label.view_as(pred)))
            incorrect =list( img for img, is_correct in pairs if not is_correct )
            img = tvutils.make_grid(incorrect)
            self.log_writer.add_image("False Pred", img, self.metrics.metrics['batch_count'])

    def on_after_epoch(self, *args,**kwargs):
        epoch = self.metrics.metrics['epoch_count']
        loss = self.metrics.metrics['loss_epoch']
        acuracy = self.metrics.metrics.get('accuracy_epoch', -1)

        if self.log_settings.basic:
            self.log_writer.add_scalar('loss', loss, epoch)
            self.log_writer.add_scalar('accuracy', acuracy, epoch)

        # histogram logging
        if self.log_settings.param_histo_freq > 0 and self.log_settings.param_histo_next_epoch == epoch:
            for name, m in self.model.named_modules():
                if isinstance(m, nn.Conv2d) or isinstance(m, nn.BatchNorm2d) or isinstance(m, nn.Linear):
                    self.log_writer.add_histogram(name + "/" + "weight", m.weight, epoch)
                    self.log_writer.add_histogram(name + "/" + "bias", m.bias, epoch)           
            self.log_settings.param_histo_next_epoch += self.log_settings.param_histo_freq

        print("[{}] Epoch: {}, loss: {:.4f}, accuracy: {:.4f}, Time: {:.2f}".format(self.run_name, epoch, 
            loss, acuracy, self.metrics.metrics['epoch_time']))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="regim",
    version="0.1.0",
    author="Shital Shah",
    author_email="sytelus@gmail.com",
    description="PyTorch Train Test Regiment Pipeline",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sytelus/regim",
    packages=setuptools.find_packages(),
	license='MIT',
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
    install_requires=[
          'torch', 'torchvision', 'longview', 'tensorboardX', 'numpy'
    ]
)


import torch
import argparse


class Config:
    class CommonEpochConfig:
        def __init__(self, batch_size):
            self.batch_size = batch_size
    class TrainEpochConfig(CommonEpochConfig):
        def __init__(self, batch_size):
            super(Config.TrainEpochConfig, self).__init__(batch_size)
            self.lr = 0.01
            self.momentum = 0.5
    class TestEpochConfig(CommonEpochConfig):
        def __init__(self, batch_size):
            super(Config.TestEpochConfig, self).__init__(batch_size)
            self.batch_size = batch_size

    def __init__(self, train_batch_size, test_batch_size):
        self.train_config = Config.TrainEpochConfig(train_batch_size)
        self.test_config = Config.TestEpochConfig(test_batch_size)
        self.epochs = 100
        self.seed = 1


    @staticmethod
    def from_args(train_batch_size=64, test_batch_size=1000, rand_seed=42, 
                  no_cuda_train=False, no_cuda_test=False):

        parser = argparse.ArgumentParser(description='PyTorch Deep Learning Pipeline')
        parser.add_argument('--batch-size', type=int, default=train_batch_size, metavar='N', #64
                            help='input batch size for training')
        parser.add_argument('--test-batch-size', type=int, default=test_batch_size, metavar='N',
                            help='input batch size for testing')
        parser.add_argument('--epochs', type=int, default=100, metavar='N',
                            help='number of epochs to train')
        parser.add_argument('--lr', type=float, default=0.01, metavar='LR', #0.01
                            help='learning rate')
        parser.add_argument('--momentum', type=float, default=0.5, metavar='M',
                            help='SGD momentum')
        parser.add_argument('--no-cuda-train', action='store_true', default=no_cuda_train,
                            help='disables CUDA training')
        parser.add_argument('--no-cuda-test', action='store_true', default=no_cuda_test,
                            help='disables CUDA testing')
        parser.add_argument('--seed', type=int, default=rand_seed, metavar='S',
                            help='random seed (default: 42)')

        args = parser.parse_args()

        config = Config(args.batch_size, args.test_batch_size)

        config.train_config.use_cuda = not args.no_cuda_train and torch.cuda.is_available()
        config.train_config.device = torch.device("cuda" if config.train_config.use_cuda else "cpu")
        config.train_config.lr = args.lr
        config.train_config.momentum = args.momentum

        config.test_config.use_cuda = not args.no_cuda_test and torch.cuda.is_available()
        config.test_config.device = torch.device("cuda" if config.test_config.use_cuda else "cpu")

        config.epochs = args.epochs

        if config.seed is not None:
            torch.manual_seed(config.seed)

        return config



import timeit

class Metrics:
    def __init__(self, callbacks):
        self.metrics = {}

        self.metrics['epoch_count'] = 0
        self.metrics['loss_sum'] = 0
        self.metrics['loss_epoch'] = 0
        self.metrics['batch_count'] = 0
        self.metrics['batch_time'] = 0.
        self.metrics['epoch_time'] = 0.
        
        callbacks.before_batch.subscribe(self.on_before_batch)
        callbacks.before_epoch.subscribe(self.on_before_epoch)
        callbacks.after_batch.subscribe(self.on_after_batch)
        callbacks.after_epoch.subscribe(self.on_after_epoch)

    def on_before_epoch(self, *args,**kwargs):
        self.epoch_start_time = timeit.default_timer()
        self.metrics['loss_sum'] = 0
        self.metrics['batch_count'] = 0
        self.metrics['epoch_count'] += 1

    def on_before_batch(self, *args,**kwargs):
        self.batch_start_time = timeit.default_timer()
        self.metrics['batch_count'] += 1

    def on_after_batch(self, test_train, input, label, output, loss):
        self.metrics['batch_time'] = timeit.default_timer() - self.epoch_start_time
        self.metrics['loss_sum'] += loss * len(input)

    def on_after_epoch(self, test_train, dataset):
        self.metrics['epoch_time'] = timeit.default_timer() - self.epoch_start_time
        self.metrics['loss_epoch'] = self.metrics['loss_sum'] / len(dataset)

class ClassificationMetrics(Metrics):
    def on_before_epoch(self, *args,**kwargs):
        super(ClassificationMetrics, self).on_before_epoch(*args,**kwargs)
        self.metrics['correct_sum'] = 0
        self.metrics['accuracy_epoch'] = 0

    def on_after_batch(self, test_train, input, label, output, loss):
        super(ClassificationMetrics, self).on_after_batch(test_train, input, label, output, loss)
        pred = output.max(1, keepdim=True)[1] # get the index of the max log-probability
        self.metrics['correct_sum'] += pred.eq(label.view_as(pred)).sum().item()

    def on_after_epoch(self, test_train, dataset):
        super(ClassificationMetrics, self).on_after_epoch(test_train, dataset)
        self.metrics['accuracy_epoch'] = self.metrics['correct_sum'] / len(dataset)


        




import torch
from torchvision import datasets, transforms

from .train_test import TrainTest
from .debug_probe import DebugProbe
from .metrics import Metrics, ClassificationMetrics

class Pipeline:
    class TaskType:
        classification = 0
        regression = 1

    def __init__(self, config):
        self.config = config

    def run(self, model, train_loader, test_loader, task_type, 
                 loss_module, optimizer, initializer=None):

        train_test = TrainTest(model, self.config, optimizer, loss_module, initializer)

        if task_type == Pipeline.TaskType.classification:
            train_metrics = ClassificationMetrics(train_test.train_callbacks)
            test_metrics = ClassificationMetrics(train_test.test_callbacks)
        elif task_type == Pipeline.TaskType.regression:
            train_metrics = Metrics(train_test.train_callbacks)
            test_metrics = Metrics(train_test.test_callbacks)
        else:
            raise NotImplementedError()

        train_probe = DebugProbe('mnist_official_pipeline', 'train', self.config.train_config, 
                                 model, train_test.train_callbacks, train_metrics)
        test_probe = DebugProbe('mnist_official_pipeline', 'test', self.config.test_config, 
                                model, train_test.test_callbacks, test_metrics)

        train_test.train_model(self.config.epochs, train_loader, test_loader)
